"""
Custom log handler for posting errors to errordite (www.errordite.com).

Dependencies: Requests (http://docs.python-requests.org)
"""
__title__ = 'errordite'
__version__ = '0.5.1'
__author__ = 'Hugo Rodger-Brown'
__license__ = 'Simplified BSD License'
__copyright__ = 'Copyright 2013 Hugo Rodger-Brown'
__description__ = 'Errordite exception logging.'

from handlers import ErrorditeHandler



"""
Basic tests for errordite log handler.

You will need a valid Errordite account in order to run these tests,
and will need to set a local environment variable ERRORDITE_TOKEN with
your Errordite account token. (This is so that the token is not stored
in the public repo.)
"""
import os
import unittest
import logging
from . import ErrorditeHandler

ERRORDITE_TOKEN = os.environ.get("ERRORDITE_TOKEN", None)
if ERRORDITE_TOKEN is None:
    raise Exception("You must set environment variable ERRORDITE_TOKEN "
        "before running these tests.")


class ErrorditeHandlerTestCase(unittest.TestCase):

    def setUp(self):
        self.handler = ErrorditeHandler(ERRORDITE_TOKEN)
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(self.handler)
        self.logger.level = logging.DEBUG
        self.assertEqual(len(self.logger.handlers), 1)

    def tearDown(self):
        self.logger.handlers = []

    def test_logging_no_exception_does_nothing(self):
        "Test that a log record with no exc_info doesn't do anything."
        func = self.test_logging_no_exception_does_nothing
        self.logger.debug(func.__doc__)

    def test_logging_error(self):
        "Test logging a real exception with logging.error."
        func = self.test_logging_error
        try:
            throw_exception(func.__name__)
        except CustomException:
            self.logger.error(func.__doc__)
            print ("Please check Errordite for a new issue with the "
                "messsage '%s'" % func.__doc__)

    def test_logging_exception(self):
        "Test logging a real exception with logging.exception."
        func = self.test_logging_exception
        try:
            throw_exception(func.__name__)
        except CustomException:
            self.logger.exception(func.__doc__)
            print ("Please check Errordite for a new issue with the "
                "messsage '%s'" % func.__doc__)

    def test_logging_with_complex_message(self):
        "Test the user of format string and args in logged message."
        func = self.test_logging_with_complex_message
        try:
            throw_exception(func.__name__)
        except CustomException:
            self.logger.exception("%s: %s", func.__doc__, 'ok')
            print ("Please check Errordite for a new issue with the "
                "messsage '%s'" % func.__doc__)


class CustomException(Exception):
    def __init__(self, message):
        super(CustomException, self).__init__(message)


def throw_exception(message):
    raise CustomException(message)



"""
Setup file for python-errordite.
"""
import os
from os.path import join, dirname, normpath, abspath
from setuptools import setup

# we can't import errordite as it hasn't been installed yet, so instead
# read information in as text. This will read in the __init__ file, and
# create a dict containing any lines beginning with '__' as keys, with
# whatever is after the '=' as the value,
# __desc__ = 'hello'
# would give {'desc': 'hello'}
meta = {}
for l in [line for line in tuple(open('errordite/__init__.py', 'r')) if line[:2] == '__']:
    t = l.split('=')
    meta[t[0].strip().strip('__')] = t[1].strip().strip('\'')

# allow setup.py to be run from any path
os.chdir(normpath(join(abspath(__file__), os.pardir)))

setup(
    name=meta['title'],
    version=meta['version'],
    packages=['errordite'],
    include_package_data=True,
    install_requires=['requests'],
    license=open(join(dirname(__file__), 'LICENCE.md')).read(),
    description=meta['description'],
    long_description=open(join(dirname(__file__), 'README.rst')).read(),
    url='https://github.com/yunojuno/python-errordite',
    author=meta['author'],
    author_email='hugo@rodger-brown.com',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
)



"""
Definition of ErrorditeHandler. Imported into package root.
"""
import logging
import sys
import traceback
import datetime
import platform
import requests
import json

# see https://www.errordite.com/help/senderrorwithjson for details:
ERRORDITE_API_URL = 'https://www.errordite.com/receiveerror'


class ErrorditeHandler(logging.Handler):
    """
    Log handler used to send exceptions to errordite.
    """
    def __init__(self, token):
        """
        Args:
            token: the auth token for access to the API - see errordite.com
        """
        logging.Handler.__init__(self)
        self.token = token

    def enrich_errordite_payload(self, payload, record):
        """
        Create the Errordite JSON from the log record.

        This is a separate method that can be overridden in subclasses.
        This decouples specific framework exception (e.g. Django) from
        basic exceptions that may not, for instance, include HTTP request
        information.

        If you are subclassing this you should always call the base
        implementation first as this fills the basic information out.

        Args:
            record: the logger record passed to the emit method.

        Returns:
            JSON-serializable dictionary that can be sent to Errordite
            as the HTTP payload.
        """
        return payload

    def emit(self, record):
        """
        Sends exception info to Errordite.

        This handler will ignore the log level, and look for an exception
        within the record (as recored.exc_info) or current stack frame
        (sys.exc_info()). If it finds neither, it will simply return without
        doing anything.
        """
        if not self.token:
            raise Exception("Missing Errordite service token.")

        if record.levelname == 'EXCEPTION':
            exc_info = record.exc_info
        else:
            exc_info = sys.exc_info()

        if exc_info == (None, None, None):
            # we can't find an exception to report on, so just return
            return

        ex_type, ex_value, ex_tb = exc_info
        ex_source = traceback.extract_tb(ex_tb)[-1]

        payload = {
            "TimestampUtc": datetime.datetime.utcnow().isoformat(),
            "Token": self.token,
            "MachineName": platform.node(),
            "ExceptionInfo": {
                "Message": record.msg % record.args,
                "Source": '%s: line %s' % (ex_source[0], ex_source[1]),
                "ExceptionType": '%s.%s' % (ex_type.__module__, ex_type.__name__),
                "StackTrace": traceback.format_exc(),
                "MethodName": ex_source[2]
            }
        }
        if hasattr(record, 'version'):
            payload['Version'] = record.version

        # enrich with additional, non-core information. This may be sub-
        # classed
        payload = self.enrich_errordite_payload(payload, record)

        try:
            requests.post(
                ERRORDITE_API_URL,
                data=json.dumps(payload),
                headers={'content-type': 'application/json'}
            )
            # since we already in the logger, logging an error, there's
            # there's really nothing we can do with the response that adds
            # any value - so ignore it.
            return 'ok'

        except:
            self.handleError(record)
